#!/usr/bin/env python
# Copyright 2016 Cumulus Networks, Inc.
#
# mroute-check kernel helper

import re
import os
import subprocess
import sys

import cumulus.porttab

class Kmroute:
    def __init__(self, sip, dip, iif, oil, orig_str):
        self.sip = sip
        self.dip = dip
        self.iif = iif
        self.oil = oil
        self.orig_str = orig_str

    def __str__(self):
        return self.orig_str

class Kernel:
    _devs_supported = ('bridge', 'bond', 'vlan')

    def __init__(self):
        self.pt = cumulus.porttab.porttab()
        self.linux_ports = self.pt.get_linux_ports()
        # key: devName info: link_kind
        self.dev_kind = {}

        # pull the kernel mroute table
        # dict key: '(1.1.1.1, 239.1.1.1)', info: Kmroute
        self.mroutes = {}
        self._mroute_parse_iproute2_output()

    def __str__(self):
        ipmr_str = 'Kernel ip mroutes:\n'
        for key in self.mroutes:
            ipmr_str += '%s\n' % self.mroutes[key]
        return ipmr_str

    @staticmethod
    def _get_cmd_output(cmdList, dispStr=None):
        try:
            with open(os.devnull, 'w') as fnull:
                dispStr = subprocess.check_output(cmdList, stderr=fnull)
        except subprocess.CalledProcessError:
            pass
        return dispStr

    def _mroute_install_hw_dev_ok(self, dev):
        if dev in self.linux_ports:
            return True

        # check if the device is in the cache
        kind = self.dev_kind.get(dev)
        if not kind:
            # if not in cache get and cache it
            cmd = '/bin/ip -d link show %s' % dev
            dispStr = self._get_cmd_output(cmd.split(), '')

            lines = dispStr.splitlines()
            if not lines or len(lines) < 3 or 'does not exist' in lines[0]:
                return False

            kind = lines[2].split()[0]
            self.dev_kind[dev] = kind

        return True if kind in self._devs_supported else False

    def _mroute_install_hw_ok(self, mroute):
        if not self._mroute_install_hw_dev_ok(mroute.iif):
            return False

        for oif in mroute.oil:
            if not self._mroute_install_hw_dev_ok(oif):
                return False

        return True

    def _mroute_key2sip_and_dip(self, key):
        pat=re.compile('\((?P<sip>\S+), (?P<dip>\S+)\)')
        obj = pat.match(key)
        if obj:
            return obj.group('sip'), obj.group('dip')
        return None, None

    def _mroute_parse_iproute2_output(self):
        cmd = '/bin/ip mroute show'
        dispStr = self._get_cmd_output(cmd.split(), '')

        # sample format -
        # '(0.0.0.0, 239.1.1.2)             Iif: swp2s3     Oifs: swp2s3 br0 '
        pat = re.compile('Iif: (?P<iif>\S+)')
        for line in dispStr.splitlines():
            key = line.split('Iif:')[0].strip()
            sip, dip = self._mroute_key2sip_and_dip(key)
            if not sip or not dip:
                continue
            oil = []
            if 'Oifs:' in line:
                oil = line.split('Oifs:')[1].split()
            obj = pat.search(line)
            if obj:
                iif = obj.group('iif')
            else:
                iif = 'unk'
            mroute = Kmroute(sip, dip, iif, oil, line)
            if self._mroute_install_hw_ok(mroute):
                self.mroutes[key] = mroute

    def get_mroutes(self):
        # return a set of mroute keys
        return set(self.mroutes)

    def get_mroute_iif_and_oil(self, key):
        mroute = self.mroutes.get(key)
        if mroute:
            return mroute.iif, mroute.oil
        return None, None

    def dump_mroutes(self, filter_mroutes=None):
        if not filter_mroutes:
            for key in self.mroutes:
                sys.stdout.write('%s\n' % self.mroutes[key])
            return

        for key in filter_mroutes:
            mroute = self.mroutes.get(key)
            if mroute:
                sys.stdout.write('%s\n' % mroute)
